#  Coordinate module for timekeeping and Earth reference frames.
#
#  Author: Stephen Leroy (sleroy@aer.com)
#  Version: 2.3.0
#  Date: 15 February 2024
#
#  This package contains two classes for use in converting between time standards:
#  Calendar and Time. The Calendar class uses calendar notation to define a time
#  that is not anchored in a time standard. The Time class categorizes a Calendar
#  instance by a specific time standard, GPS time, TAI time, or UTC time, and
#  performs all conversions. For example:
#
#  >>> calendar_time = Calendar( year=2019, month=12, day=18, hour=16, minute=36 )
#  >>> standard_time = Time( gps=calendar_time )
#
#  which defines a variable standard_time that corresponds to the date and time
#  prescribed in calendar_time as a GPS time. The standard_time also contains
#  the TAI time and UTC time corresponding the GPS time defined in calendar_time.
#
#  The "calendar" method of a Time instance produces a Calendar instance of any
#  of TAI time, GPS time, or UTC time, depending on the argument. For example,
#
#  >>> utc_calendar_time = standard_time.calendar( "utc" )
#  >>> tai_calendar_time = standard_time.calendar( "tai" )
#
#  The "gpstime" method of a Time instance produces a 5-element tuple giving
#  the GPS week, day of week, hour, minute, and second corresponding to the
#  Time instance.
#
#  >>> gps_week, day_of_week, hour, minute, second = standard_time.gpstime()
#

#  Revisions.
#  25 March 2020:
#    * Increase precision of time-keeping by redefining t1900 class by counting
#      seconds and fractional seconds separately.
#    * Update leap-second service get_leapseconddata by checking if the existing
#      file is beyond expiration.
#
#  14 April 2020:
#    * Time() returns an instance representing the current time if called
#      without arguments.
#
#  17 November 2021: 
#    * Class Calendar will be initialized by year, month, day, etc. by default. 
#    * Add datetime and isoformat methods to class Calendar. 
#    * Remove obstime method from class Time. 
#    * Fix bug in Calendar class that rendered an incorrect day-of-year 
#      computation. 
#
#  17 October 2023: 
#    * Update the site from where the leap second data are downloaded. The 
#      IERS site has been discontinued. The new table of leap seconds is 
#      updated to https://cdf.gsfc.nasa.gov/html/CDFLeapSeconds.txt. 
#
#  12 February 2024: 
#    * Fully functional Time.juliandate capability. Linked with astropy. 
#
#  13 February 2024: 
#    * Calendar.isoformat is a tunnel to datetime.datetime.isoformat. 
#
#  15 February 2024: 
#    * Allow the Time class to accept datetime.datetime and numpy.datetime64
#      objects to the tai, gps, and utc keywords at initialization. 
#

import numpy as np
import requests
import os
import re
import pathlib
import datetime as Datetime
from astropy.time import Time as astropyTime

#  Global parameters.

pi = np.pi
rads = pi/180

#  Exception handling.

class Error(Exception):
    pass


################################################################################
#                              Class Calendar
################################################################################

class CalendarError(Error):
    def __init__(self,expression,message):
        self.expression = expression
        self.message = message

class t1900Error(Error):
    def __init__(self,expression,message):
        self.expression = expression
        self.message = message

class t1900():
    def __init__(self,ix,fx=0.0):
        """Instance of class t1900, meant to count the number of (Gregorian) seconds elapsed
        since 1 Jan 1900 at 00:00 UTC without accounting for leap seconds. It splits the
        number of seconds into an integer part and a fractional part."""

        self._scale = 2**64

        if not isinstance(ix,int) and not isinstance(ix,np.int32) and not isinstance(ix,np.int64) :
            raise t1900Error( "InvalidArgument", "First argument must be an int." )
        self._integerSeconds = int( ix )

        if not isinstance(fx,float) and not isinstance(fx,np.float32) and not isinstance(fx,np.float64):
            raise t1900Error( "InvalidArgument", "Second argument must be a float." )
        self._fractionalSeconds = int( fx * self._scale + 0.5 )

    def seconds(self):
        """Return a two-tuple, the first element containing the integer number of
        seconds and the second element containing the fractional number of seconds
        as a float."""

        return self._integerSeconds, self._fractionalSeconds / self._scale

    def __int__(self):
        return self._integerSeconds

    def __eq__(self,x):
        return ( self._integerSeconds == x._integerSeconds and self._fractionalSeconds >= x._fractionalSeconds )

    def __gt__(self,x):
        return self._integerSeconds > x._integerSeconds \
            or ( self._integerSeconds == x._integerSeconds and self._fractionalSeconds > x._fractionalSeconds )

    def __ge__(self,x):
        return self._integerSeconds > x._integerSeconds \
            or ( self._integerSeconds == x._integerSeconds and self._fractionalSeconds >= x._fractionalSeconds )

    def __lt__(self,x):
        return self._integerSeconds < x._integerSeconds \
            or ( self._integerSeconds == x._integerSeconds and self._fractionalSeconds < x._fractionalSeconds )

    def __le__(self,x):
        return self._integerSeconds < x._integerSeconds \
            or ( self._integerSeconds == x._integerSeconds and self._fractionalSeconds <= x._fractionalSeconds )

    def __add__(self,x):
        if not isinstance( x, int ) and not isinstance( x, float ):
            raise t1900Error( "InvalidOperation", "Addition requires adding a t1900 to an int or float.." )
        ix, fx = int( x ), int( ( x - int(x) ) * self._scale )
        if fx + self._fractionalSeconds >= self._scale:
            return t1900( self._integerSeconds + ix + 1, ( self._fractionalSeconds + fx - self._scale )/self._scale )
        else:
            return t1900( self._integerSeconds + ix, ( self._fractionalSeconds + fx )/self._scale )

    def __sub__(self,x):
        if isinstance( x, int ) + isinstance( x, float ) + isinstance( x, t1900 ) != 1:
            raise t1900Error( "InvalidOperation", "Subtraction requires subtracting a float, int, " + \
                "or t1900 instance from a t1900 instance." )

        if isinstance( x, float ):
            ix, fx = int( x ), int( ( x - int(x) ) * self._scale )
            if self._fractionalSeconds >= fx:
                return t1900( self._integerSeconds - ix, ( self._fractionalSeconds - fx )/self._scale )
            else:
                return t1900( self._integerSeconds - ix - 1, ( self._fractionalSeconds - fx + self._scale )/self._scale )

        if isinstance( x, int ):
            return t1900( self._integerSeconds - x, self._fractionalSeconds/self._scale )

        if isinstance( x, t1900 ):
            if x._fractionalSeconds < self._fractionalSeconds:
                return ( self._integerSeconds - x._integerSeconds ) + \
                    ( self._fractionalSeconds - x._fractionalSeconds ) / self._scale
            else:
                return ( self._integerSeconds - x._integerSeconds - 1 ) + \
                    ( self._fractionalSeconds - x._fractionalSeconds + self._scale) / self._scale

#  Magic methods. 

    def __iadd__(self,x):
        if not isinstance( x, int ) and not isinstance( x, float ):
            raise t1900Error( "InvalidOperation", "Addition requires adding a t1900 to an int or float.." )
        ix, fx = int( x ), int( ( x - int(x) ) * self._scale )
        self._integerSeconds += ix
        self._fractionalSeconds += fx
        if self._fractionalSeconds > self._scale:
            self._integerSeconds += 1
            self._fractionalSeconds -= self._scale
        return self

    def __isub__(self,x):
        if not isinstance( x, int ) and not isinstance( x, float ):
            raise t1900Error( "InvalidOperation", "Addition requires adding a t1900 to an int or float.." )
        ix, fx = int( x ), int( ( x - int(x) ) * self._scale )
        self._integerSeconds -= ix
        self._fractionalSeconds -= fx
        if self._fractionalSeconds < 0:
            self._integerSeconds -= 1
            self._fractionalSeconds += self._scale
        return self


class Calendar():
    """This class provides methods for calendar operations, converting between calendar
    notation (year, month, day, etc.) and t1900 time, the latter of which counts
    seconds from 1 Jan 1900 at 00:00:00 (with 86400 seconds in every day). It also
    keeps track of the day-of-year and day-of-week."""

    def __init__(self, year=None, month=None, day=None, hour=None, minute=None, second=None, 
                 t1900t=None, system=None ):
        """Initialization takes either an instance of t1900 as a single argument or year,
        month, day in order to form a Calendar object. In the latter case, hour, minute
        and second can also be given optionally."""

        self.year = 0
        self.month = 0
        self.day = 0
        self.hour = 0
        self.minute = 0
        self.second = 0.0
        self.t1900 = None
        self.dow = 0			# 0 for Sunday, 1 for Monday, etc.
        self.doy = 0			# 1 for January 1, etc.
        self.datetimestring = ''
        self.datestring = ''

        if system is None: 
            self.system = None
        elif system.lower() in [ "utc", "tai", "gps" ]: 
            self.system = system.lower()
        else: 
            self.system = None

        #  Work with t1900.

        if t1900t is not None:

        #  Check the argument.

            if not isinstance( t1900t, t1900 ):
                raise CalendarError( "InvalidArguments", "Argument must be a t1900 class instance" )

            #  Check if keyword arguments were sent.

            if not ( year is None and month is None or day is None \
                or hour is None or minute is None or second is None ):
                raise CalendarError( "InvalidArguments", "Date specified multiple ways" )

            self.t1900 = t1900t
            self.get_date()

        else:

            if year is None or month is None or day is None:
                raise CalendarError( "InvalidArguments", "Must specify at least year, month, and day" )

            self.year = int( year )
            self.month = int( month )
            self.day = int( day )
            if hour is not None: self.hour = int(hour)
            if minute is not None: self.minute = int(minute)
            if second is not None: self.second = float(second)

            self.get_t1900()

        self.compose_strings()

    #  Compute t1900 for New Years Day given a year.

    def newyearsday_t1900(self,year):
        t = t1900( ( year*365 + int((year-1)/4) - int((year-1)/100) + int((year-1)/400) - 693960 ) * 86400 )
        return t

    #  Determine whether or not it is a leap year.

    def ndays(self,year):
        if ( ( year%4 == 0 ) and ( year%100 != 0 ) or ( year%400 == 0 ) ):
            return 366
        else:
            return 365

    #  Get the date from t1900

    def get_date(self):

        #  Determine year.

        res = self.t1900.seconds()[0]

        year = int( res / 365.0 / 86400.0 ) + 1900
        et1900 = self.newyearsday_t1900(year)
        while ( et1900 > self.t1900 ):
            year -= 1
            et1900 = self.newyearsday_t1900(year)
        res -= et1900.seconds()[0]
        self.year = year

        #  Determine day-of-year.

        self.doy = int( res / 86400.0 ) + 1

        #  Determine month.

        if ( self.ndays(year) == 365 ):
            days = [ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ]
        else:
            days = [ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ]

        for month in range(12):
            if ( res < days[month] * 86400.0 ): break
            res -= days[month] * 86400.0
        self.month = month + 1

        #  Determine day, hour, minute, second, day-of-week.

        self.day = int( res / 86400.0 ) + 1
        res -= ( self.day - 1 ) * 86400.0
        self.hour = int( res / 3600.0 )
        res -= self.hour * 3600.0
        self.minute = int( res / 60.0 )
        res -= self.minute * 60.0
        self.second = int( res ) + self.t1900.seconds()[1]
        self.dow = ( int(self.t1900.seconds()[0]/86400.0) + 1 ) % 7		# Check this.

        #  Compose date strings.

    def compose_strings(self):
        months = 'Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec'.split()
        self.datetimestring = '{2:02d} {1} {0} {3:02d}:{4:02d}:{5:02d}'.format( self.year,
            months[self.month-1], self.day, self.hour, self.minute, int(self.second) )
        self.datestring = '{2:02d} {1} {0}'.format( self.year,
            months[self.month-1], self.day, self.hour, self.minute, self.second )

    #  Get t1900 from the date.

    def get_t1900(self):

        et1900 = self.newyearsday_t1900(self.year)
        self.t1900 = et1900 + 0
        if ( self.ndays(self.year) == 365 ):
            days = [ 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ]
        else:
            days = [ 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 ]
        for month in range(self.month-1): self.t1900 += days[month] * 86400
        self.t1900 += (self.day-1)*86400 + self.hour*3600 + self.minute*60
        self.t1900 += self.second
        self.doy = int( (self.t1900.seconds()[0] - et1900.seconds()[0])/86400 ) + 1
        self.dow = ( int(self.t1900.seconds()[0]/86400) + 1 ) % 7		# Check this.

    #  Output a datetime.datetime instance.

    def datetime( self ):
        """Output an instance of datetime.datetime."""

        microseconds = int( ( self.second - int(self.second) ) * 1.0e6 )
        dt = Datetime.datetime( year=self.year, month=self.month, day=self.day,
                hour=self.hour, minute=self.minute, second=int(self.second), 
                microsecond=microseconds )

        return dt

    #  Output an isotime instance. 

    def isoformat( self, **kwargs ): 
        """Output an string in ISO time format."""

        dt = self.datetime()
        ret = dt.isoformat( **kwargs )
        return ret

    #  Magic methods. 

    def __iadd__(self,x):
        self.t1900 += x
        self.get_date()
        self.compose_strings()
        return self

    def __isub__(self,x):
        self.t1900 -= x
        self.get_date()
        self.compose_strings()
        return self


################################################################################
#                              Class Time
################################################################################

class TimeError(Error):
    def __init__(self,expression,message):
        self.expression = expression
        self.message = message

class Time():
    """This class provide for conversions between TAI (atomic time), UTC, and GPS time.
    It is also helpful for generating GPS format time and Julian day."""

    def __init__(self,tai=None,gps=None,utc=None,leapsecondfile=None,system=None):
        """Create a time-conversion object. "leapsecondfile" should point toward a leap
        second list as obtained from https://www.ietf.org/timezones/data/leap-seconds.list.
        Its default behavior is to obtain the list by http request. "system" should be
        either 1 or 2, defining how UTC is computed from TAI. If system 1, then the time
        23:59:59 UTC is repeated immediately before the leap day. If system 2, then the
        time 23:59:60 is included and proceeds to 0:00:00 on the next day for a leap day
        included. The default is system 2.

        Only one of "tai", "gps", or "utc" should be given. The object given should be
        either an instance of class t1900 or an instance of class Calendar. In the case 
        of "gps", the value can be GPS seconds as a float or a floatable number."""

        self.tai = None
        self.utc = None
        self.gps = None

        #  Get leap second data.

        global __LEAPSECONDDATA__
        if __LEAPSECONDDATA__ is None : get_leapseconddata( leapsecondfile )
        self.leapseconddata = __LEAPSECONDDATA__

        #  Define GPS time origin.

        self.gps0 = t1900( 2524953600, 0.0 )

        #  Define the UTC leap second system.

        if system in {1,2}:
            self.system = system
        else:
            self.system = 2

        if tai is None and gps is None and utc is None:
            now = Datetime.datetime.now( tz=Datetime.timezone.utc )
            utc = Calendar( year=now.year, month=now.month, 
                    day=now.day, hour=now.hour, minute=now.minute, 
                    second=now.second+now.microsecond*1.0e-6 )

        #  Be sure that one of tai, gps, or utc is given.

        if int( tai is None ) + int( gps is None ) + int( utc is None ) != 2:
            raise TimeError( "InvalidArguments", "Must specify only one of tai, gps, or utc" )

        #  Convert tai to gps and utc time. UTC requires interpretation of leap seconds.

        if tai is not None:

            if isinstance(tai,t1900): 
                self.tai = tai
            elif isinstance(tai,Calendar): 
                self.tai = tai.t1900
            elif isinstance(tai,Datetime.datetime): 
                dt = tai
                cal = Calendar( year=dt.year, month=dt.month, day=dt.day, 
                               hour=dt.hour, minute=dt.minute, second=dt.second+dt.microsecond*1.0e-6 )
                self.tai = cal.t1900
            elif isinstance(tai,np.datetime64): 
                dt = Datetime.datetime.fromisoformat( str(tai) )
                cal = Calendar( year=dt.year, month=dt.month, day=dt.day, 
                               hour=dt.hour, minute=dt.minute, second=dt.second+dt.microsecond*1.0e-6 )
                self.tai = cal.t1900
            else: 
                raise TimeError( "InvalidArguments", "tai must be an instance of class t1900, " + \
                        "timestandards.Calendar, datetime.datetime, or numpy.datetime64" )

            self.gps = self.tai - 19.0
            for leapseconddatum in self.leapseconddata:
                self.utc = self.tai - leapseconddatum['leapseconds']
                if self.system==1 and self.utc >= leapseconddatum['t1900']-1: break
                if self.system==2 and self.utc >= leapseconddatum['t1900']: break

        #  Convert gps to tai and utc time. UTC requires interpretation of leap seconds.

        if gps is not None:

            if isinstance(gps,t1900): 
                self.gps = gps
            elif isinstance(gps,Calendar): 
                    self.gps = gps.t1900
            elif isinstance(gps,Datetime.datetime): 
                dt = gps
                cal = Calendar( year=dt.year, month=dt.month, day=dt.day, 
                               hour=dt.hour, minute=dt.minute, second=dt.second+dt.microsecond*1.0e-6 )
                self.gps = cal.t1900
            elif isinstance(gps,np.datetime64): 
                dt = Datetime.datetime.fromisoformat( str(gps) )
                cal = Calendar( year=dt.year, month=dt.month, day=dt.day, 
                               hour=dt.hour, minute=dt.minute, second=dt.second+dt.microsecond*1.0e-6 )
                self.gps = cal.t1900
            else: 
                try: 
                    self.gps = self.gps0 + float( gps )
                except: 
                    raise TimeError( "InvalidArguments", "gps must be an instance of class t1900, " + \
                        "timestandards.Calendar, datetime.datetime, numpy.datetime64, or a " + \
                        "floatable object." )

            self.tai = self.gps + 19.0
            for leapseconddatum in self.leapseconddata:
                self.utc = self.tai - leapseconddatum['leapseconds']
                if self.system==1 and self.utc >= leapseconddatum['t1900']-1: break
                if self.system==2 and self.utc >= leapseconddatum['t1900']: break

        #  Convert utc to gps and tai time. Because of leap seconds, at the instances of
        #  leap seconds, the conversions are ill defined.

        if utc is not None:
            if isinstance(utc,t1900): 
                self.utc = utc
            elif isinstance(utc,Calendar): 
                self.utc = utc.t1900
            elif isinstance(utc,Datetime.datetime): 
                dt = utc
                cal = Calendar( year=dt.year, month=dt.month, day=dt.day, 
                               hour=dt.hour, minute=dt.minute, second=dt.second+dt.microsecond*1.0e-6 )
                self.utc = cal.t1900
            elif isinstance(utc,np.datetime64): 
                dt = Datetime.datetime.fromisoformat( str(utc) )
                cal = Calendar( year=dt.year, month=dt.month, day=dt.day, 
                               hour=dt.hour, minute=dt.minute, second=dt.second+dt.microsecond*1.0e-6 )
                self.utc = cal.t1900
            else: 
                raise TimeError( "InvalidArguments", "utc must be an instance of class t1900, " + \
                        "timestandards.Calendar, datetime.datetime, or numpy.datetime64" )

            for leapseconddatum in self.leapseconddata:
                if self.utc >= leapseconddatum['t1900']:
                    self.tai = self.utc + leapseconddatum['leapseconds']
                    break
            if self.tai is None: self.tai = self.utc + self.leapseconddata[-1]['leapseconds'] - 1
            self.gps = self.tai - 19.0

    def juliandate(self):
        """Compute the Julian date."""

        #  Use astropy.Time. 

        atime = astropyTime( [self.calendar("utc").isoformat()], format='isot', scale='utc' )[0]
        jd = atime.jd

        return jd

    def gpsweekday(self):
        """Generate a two-tuple of integers. The first elemtn is the GPS week, the second
        element is the day-of-week."""

        gpsdays = int( ( self.gps - self.gps0 ) / 86400 )
        gpsweek = int( gpsdays / 7 )
        dow = gpsdays % 7

        return gpsweek, dow

    def calendar(self, arg):
        """Generate a Calendar instance of "utc", "tai", or "gps", which are the only valid
        values of arg."""

        args = { 'system': arg }
        if arg == 'utc': 
            args.update( { 't1900t': self.utc } )
        elif arg == 'tai': 
            args.update( { 't1900t': self.tai } )
        elif arg == 'gps': 
            args.update( { 't1900t': self.gps } )
        else: 
            raise TimeError( "InvalidArgument", "Argument must be either utc, tai or gps" )

        ret = Calendar( **args )

        return ret

    def gpstime(self):
        """Generate a tuple of ( gps_week, day_of_week, hour, minute, second ) corresponding
        to GPS time."""

        gps0 = self.gps0.seconds()[0]
        if self.gps.seconds()[0] >= gps0:
            gpsweek = int( ( self.gps.seconds()[0] - gps0 ) / ( 7 * 86400 ) )
            cal = self.calendar("gps")
            ret = ( gpsweek, cal.dow, cal.hour, cal.minute, cal.second )
        else:
            ret = None

        return ret

    #  Magic (operator) methods. The following magic methods are permitted:
    #
    #    ">", ">=", "<", "<=", all of which are based on comparisons of tai time.
    #
    #    "+", which can only add a float to a Time; the float having units of
    #        seconds.
    #
    #    "-", which takes the differences between two instances of Time and
    #        returns a float, which has units of seconds.
    #
    #    "+=", which functions as an increment operator, and can only take a
    #        float with units of seconds, returning another instance of Time.
    #
    #    "-=", which functions as a decrement operator, and con only take a
    #        float with units of seconds, returning another instance of Time.
    #

    def __eq__(self,x):
        return self.tai == x.tai

    def __gt__(self,x):
        return self.tai > x.tai

    def __ge__(self,x):
        return self.tai >= x.tai

    def __lt__(self,x):
        return self.tai < x.tai

    def __le__(self,x):
        return self.tai <= x.tai

    def __add__(self,x):
        new = Time( tai=self.tai+x, system=self.system )
        return new

    def __sub__(self,x):
        if isinstance(x,Time):
            new = self.tai - x.tai
        else:
            new = Time( tai=self.tai-x, system=self.system )
        return new

    def __iadd__(self,x):
        new = Time( tai=self.tai+x, system=self.system )
        return new

    def __isub__(self,x):
        new = Time( tai=self.tai-x, system=self.system )
        return new

    def __repr__(self): 
        x = f'<timestandard.Time object, "{self.calendar("utc").isoformat(timespec="seconds")} UTC">'
        return x


################################################################################
#  Service functions for leap second data.
################################################################################

class LeapSecondError(Error):
    def __init__(self,expression,message):
        self.expression = expression
        self.message = message


def download_leapsecondfile( leapsecondfile ):

    #  Download leap-second file.

    r = requests.get( "https://cdf.gsfc.nasa.gov/html/CDFLeapSeconds.txt" )
    if r.status_code != 200:
        raise LeapSecondError( "InternetFileNotFound", "Cannot retrieve leap second data over internet" )
    f = open( leapsecondfile, "w" )
    f.write( r.text )
    f.close()


def get_leapseconddata( leapsecondfile=None ):

    #  Get leap second data. It is either in file previously downloaded from
    #  NASA GSFC or it is retrieved from GSFC. 

    download = True

    #  Establish the leap-second filename.

    if leapsecondfile is None:
        if os.path.isdir( "/pub" ):
            leapsecondfile = "/pub/.gsfcleapseconds"
        else:
            leapsecondfile = str( pathlib.Path.home() ) + "/.gsfcleapseconds"

    #  Do we need to get a new leap-second file? Yes, if it doesn't exist in the
    #  given path, if it doesn't have an expiration date written in it, or if it
    #  past the expiration date.

    if os.path.isfile( leapsecondfile ):
        stat = os.stat( leapsecondfile )
        cdatetime = Datetime.datetime.fromtimestamp( stat.st_mtime, tz=Datetime.timezone.utc )
        now = Datetime.datetime.now( tz=Datetime.timezone.utc )
        ym_cdatetime = cdatetime.year + 0.5 * int( (cdatetime.month-1) / 6.0 )
        ym_now = now.year + 0.5 * int( (now.month-1) / 6.0 )
        download = ( ym_now > ym_cdatetime )

    #  Get the leap-second file if necessary.

    if download:
        download_leapsecondfile( leapsecondfile )

    f = open( leapsecondfile, "r" )
    alllines = f.readlines()
    f.close()

    #  Retain only the lines with useful leap-second data.

    lines = [ line.strip() for line in alllines if line != "" and not re.search("^;",line) ]

    #  Catalog leap second data. t1900 is the number of UTC seconds since 1 Jan 1900.

    leapseconddata = []
    for line in lines:
        strs = re.split( "\s+", line )[0:4]
        year, month, day, seconds = int(strs[0]), int(strs[1]), int(strs[2]), float( strs[3] )
        t1900i = Calendar( year=year, month=month, day=day ).t1900
        leapseconddata.append( { 't1900':t1900i, 'leapseconds':seconds } )
    leapseconddata.reverse()

    global __LEAPSECONDDATA__
    __LEAPSECONDDATA__ = leapseconddata


#  Initialize global variables.

global __LEAPSECONDDATA__
__LEAPSECONDDATA__ = None




"""collocation.py

Author: Stephen Leroy (sleroy@aer.com)
Version: 
Date: March 13, 2024

This module establishes two classes to handle radio occultation--nadir scanner 
collocation data: class Collocation and class CollocationList. In addition it 
provides a method to be used internally that generates xarray DataArrays that 
handle fill values properly.

class Collocation:
    Define a radio occultation - scanner data collocation. It contains all 
    of the metadata necessary to identify the individual soundings. It also 
    contains methods to fill extract the actual data/observations associated 
    with the occultation and nadir scanner soundings.

class CollocationList: 
    Define a list of instances of Collocation. It inherits the list class and 
    adds methods to form unions and intersections of instances of 
    CollocationList. It also includes methods to sort the list and write its 
    contents to an output NetCDF file."""


#  Imports. 

import os, json
import numpy as np
import xarray
import netCDF4 
from awsgnssroutils.database import OccList
from .timestandards import Time
from .nadir_satellite import NadirSatelliteInstrument, ScanMetadata
from .constants_and_utils import masked_dataarray

#  Exception handling. 

class Error( Exception ): 
    pass

class collocationError( Error ): 
    def __init__( self, message, comment ): 
        self.message = message
        self.comment = comment

#  Internal. 

__version__ = "1.0.0"

#  Parameters. 

fill_value = -1.0e20


#  Collocation class definition. 

class Collocation(): 
    """A class that defines an existing collocation. Its attributes are 

    occultation -> An OccList of size; it contains all metadata associated with 
            the collocated occultation. 

    nadir_satellite -> An instance of NadirSatelliteInstrument that contains 
            information on satellite and instrument name along with information on 
            the satellite's TLE's. 

    longitude -> Longitude (degrees east) of the collocated nadir-scanner sounding. 

    latitude -> Latitude (degrees north) of the collocated nadir-scanner sounding. 

    time -> An instance of timestandards.Time of the collocated nadir-scanner 
            sounding; this can be approximate value if iscan, ifootprint, or 
            scan_metadata is not provided.

    scan_metadata -> An instance of ScanMetadata containing metadata on the nadir-scanner 
            observations; it is accessed when retrieve measurement data. It 
            can be an approximate value if iscan, ifootprint, or scan_metadata is not provided.

    scan_angle -> The scan angle (degrees) of the collocated nadir-scanner sounding; if it 
            is not provided, then it will be impossible to infer iscan and ifootprint. 

    iscan -> Integer (zero-based) pointing to the scan index of scan_metadata defining the 
            collocated nadir-scanner sounding. 

    ifootprint -> Integer (zero-based) pointing to the footprint index of scan_metadata 
            defining the collocated nadir-scanner sounding. 

    Note the the rotation-collocation method can determine if an occultation-nadir-scanner 
    collocation exists, but it can only give approximate pointers (iscan, ifootprint) to 
    the actual collocated nadir-scanner sounding. The actual pointers must be determined 
    by estimation and refinement given approximate values of time and scan_angle. 
    """

    def __init__( self, occultation, nadir_satellite, longitude=None, latitude=None, time=None, 
                 scan_metadata=None, scan_angle=None, iscan=None, ifootprint=None ): 

        #  Check input. 

        integer_classes = [ int, np.byte, np.int8, np.int16, np.int32, np.int64 ] 

        if not isinstance( occultation, OccList ): 
            raise collocationError( "InvalidArgument", "First argument must be an instance of OccList" )
        elif occultation.size != 1: 
            raise collocationError( "InvalidArgument", "Occultation argument must be an OccList of size 1" )

        if not isinstance( nadir_satellite, NadirSatelliteInstrument ): 
            raise collocationError( "InvalidArgument", "Second argument must be an instance of NadirScanner" )

        x = longitude
        if x is not None: 
            if not isinstance(x,float) and not isinstance(x,np.float32) and not isinstance(x.np.float64): 
                raise collocationError( "InvalidArgument", "longitude must be a float-type object" )

        x = latitude
        if x is not None: 
            if not isinstance(x,float) and not isinstance(x,np.float32) and not isinstance(x.np.float64): 
                raise collocationError( "InvalidArgument", "latitude must be a float-type object" )

        x = time
        if x is not None: 
            if not isinstance(x,Time): 
                raise collocationError( "InvalidArgument", "time must be an instance of timestandards.Time" )

        x = scan_metadata
        if x is not None: 
            if not isinstance(x,ScanMetadata): 
                raise collocationError( "InvalidArgument", "scan_metadata must be an instance of ScanMetadata" )

        x = scan_angle
        if x is not None: 
            if not isinstance(x,float) and not isinstance(x,np.float32) and not isinstance(x.np.float64): 
                raise collocationError( "InvalidArgument", "scan_angle must be a float-type object" )

        x = iscan
        if x is not None: 
            if not any( [ isinstance( x, dtype ) for dtype in integer_classes ] ): 
                raise collocationError( "InvalidArgument", "iscan must be an integer-type object" )
            if iscan < 0: 
                raise collocationError( "InvalidArgument", "iscan must be greater than or equal to 0" )

        x = ifootprint
        if x is not None: 
            if not any( [ isinstance( x, dtype ) for dtype in integer_classes ] ): 
                raise collocationError( "InvalidArgument", "ifootprint must be an integer-type object" )
            if ifootprint < 0: 
                raise collocationError( "InvalidArgument", "ifootprint must be greater than or equal to 0" )

        if scan_metadata is not None: 
            nscans, nfootprints = scan_metadata.longitudes.shape
            if iscan is not None: 
                if iscan >= nscans: 
                    raise collocationError( "InvalidArgument", f"iscan must be less than {nscans}" )
            if ifootprint is not None: 
                if ifootprint >= nfootprints: 
                    raise collocationError( "InvalidArgument", f"ifootprint must be less than {nfootprints}" )
        
        #  Define attributes. 

        self.status = "nominal"
        self.occultation = occultation
        self.nadir_satellite = nadir_satellite
        self.longitude = longitude
        self.latitude = latitude
        self.time = time
        self.scan_metadata = scan_metadata
        self.scan_angle = scan_angle
        self.iscan = iscan
        self.ifootprint = ifootprint
        self.data = None

    def refine_scanner_indices( self ): 
        """Find the actual nadir-scanner sounding that is closest in space to the 
        occultation."""

        #  Check for necessary information. 

        if self.scan_angle is None: 
            raise collocationError( "MissingData", "scan_angle must be provided in " + \
                    "order to infer iscan, ifootprint" )

        if self.time is None: 
            raise collocationError( "MissingData", "time of collocated nadir-scanner " + \
                    "sounding must be provided in order to infer iscan, ifootprint" )

        #  Get scan data if needed. 

        if self.scan_metadata is None: 

            timerange = ( self.time - 4*self.nadir_satellite.time_between_scans, 
                         self.time + 4*self.nadir_satellite.time_between_scans )

            self.scan_metadata = self.nadir_satellite.get_geolocations( timerange )

        if len( self.scan_metadata.longitudes ) == 0: 
            self.status = "no sounder data available"
            return self

        nscans, nfootprints = self.scan_metadata.longitudes.shape

        #  Establish position of occultation, position of nadir-scanner soundings. 

        lon, lat = np.deg2rad( self.occultation.values("longitude")[0] ), np.deg2rad( self.occultation.values("latitude")[0] )
        p_occ = np.array( [ np.cos(lon)*np.cos(lat), np.sin(lon)*np.cos(lat), np.sin(lat) ] )

        lons, lats = self.scan_metadata.longitudes, self.scan_metadata.latitudes
        p_sounder = np.array( [ np.cos(lons)*np.cos(lats), np.sin(lons)*np.cos(lats), np.sin(lats) ] ).transpose( (1,2,0) )

        #  Find minimum distance. Find actual ifootprint, iscan of closest nadir-scanner sounding. 

        distances = np.arccos( np.inner( p_sounder, p_occ ) )
        i = distances.argmin()
        self.iscan = int( i / nfootprints )
        self.ifootprint = i % nfootprints

        self.longitude = np.rad2deg( lons[ self.iscan, self.ifootprint ] )
        self.latitude = np.rad2deg( lats[ self.iscan, self.ifootprint ] )

        #  Done. 

        return self

    def get_data( self, ro_processing_center ): 
        """Get occultation and nadir-scanner data. 

        Arguments
        ---------
        ro_processing_center: str
            The name of the RO processing center to use as the source of occultation data

        Returns
        ---------
        A dictionary with the following keyword-value pairs: 
            occid -> occultation identifier
            occultation -> An xarray.Dataset containing radio occultation profile data
            sounder -> An xarray.Dataset containing nadir-scanner sounding data """

        #  Download occultation data file. 

        occ_files = self.occultation.download( f"{ro_processing_center}_refractivityRetrieval", silent=True )
        if len( occ_files ) != 1: 
            raise collocationError( "InvalidOccultation", 
                    f"Unable to obtain {ro_processing_center}_refractivityRetrieval for collocated occultation" )
        occ_file = occ_files[0]

        #  Initialize output dictionary. 

        ret = {}

        #  Extract occultation data. 

        d = netCDF4.Dataset( occ_file, 'r' )

        longitude_dataarray = xarray.DataArray( self.occultation.values("longitude")[0] )
        longitude_dataarray.attrs.update( { 
            'description': "Longitude of radio occultation sounding, eastward", 
            'units': "degrees" } )

        latitude_dataarray = xarray.DataArray( self.occultation.values("latitude")[0] )
        latitude_dataarray.attrs.update( { 
            'description': "Latitude of radio occultation sounding, northward", 
            'units': "degrees" } )

        bendingangle_dataarray = masked_dataarray( d.variables['bendingAngle'][:], fill_value, 
            dims=("impactParameter",) )
        bendingangle_dataarray.attrs.update( { 
            'description': "Bending angle, ionosphere calibrated, unoptimized", 
            'units': "radians" } )

        impactparameter_dataarray = masked_dataarray( d.variables['impactParameter'][:], fill_value, 
            dims=("impactParameter",) )
        impactparameter_dataarray.attrs.update( { 
            'description': "Impact parameter of ray", 
            'units': "meters" } )

        radiusofcurvature_dataarray = xarray.DataArray( d.variables['radiusOfCurvature'].getValue() )
        radiusofcurvature_dataarray.attrs.update( { 
            'description': "Local radius of curvature of the Earth", 
            'units': "meters" } )

        refractivity_dataarray = masked_dataarray( d.variables['refractivity'][:], fill_value, 
            dims=("altitude",) )
        refractivity_dataarray.attrs.update( { 
            'description': "Refractivity", 
            'units': "N-units" } )

        geopotential_dataarray = masked_dataarray( d.variables['geopotential'][:], fill_value, 
            dims=("altitude",) )
        geopotential_dataarray.attrs.update( { 
            'description': "Geopotential energy per unit mass", 
            'units': "J/kg" } )

        altitude_dataarray = masked_dataarray( d.variables['altitude'][:], fill_value, 
            dims=("altitude",) )
        altitude_dataarray.attrs.update( { 
            'description': "Altitude above mean sea-level geoid", 
            'units': "meters" } )

        ds_occultation = xarray.Dataset( { 
            'longitude': longitude_dataarray, 
            'latitude': latitude_dataarray, 
            'bendingAngle': bendingangle_dataarray, 
            'impactParameter': impactparameter_dataarray, 
            'radiusOfCurvature': radiusofcurvature_dataarray, 
            'refractivity': refractivity_dataarray, 
            'geopotential': geopotential_dataarray, 
            'altitude': altitude_dataarray } )

        ds_occultation.attrs.update( {
            'file': occ_files[0], 
            'mission': self.occultation._data[0]['mission'], 
            'transmitter': self.occultation._data[0]['transmitter'], 
            'receiver': self.occultation._data[0]['receiver'], 
            'time': self.time.calendar("utc").isoformat(timespec="seconds")+"Z" } )

        d.close()

        #  Find the actual closest nadir-scanner sounding. 

        if self.iscan is None or self.ifootprint is None: 
            self.refine_scanner_indices()

        if self.status == "nominal": 

            #  Extract sounder data. 

            self.time = self.scan_metadata.mid_times[self.iscan] 
            ds_sounder = self.scan_metadata( self.iscan, self.ifootprint, 
                    longitude=self.longitude, latitude=self.latitude, time=self.time )

        else: 

            ds_sounder = None

        #  Merge occultation and sounder data. 

        ds = { 'occultation': ds_occultation, 'sounder': ds_sounder, 
                'occid': self.occultation._data[0]['occid'] } 

        #  Done. 

        self.data = ds

        return ds

def collocation_confusion( occs, collocations_bruteforce, collocations_rotation ): 
    """Produce a confusion matrix for the rotation-collocation algorithm. The first 
    argument must be an OccList of the occultations used for finding collocated nadir 
    soundings. The second argument must be a list of Collocation instances obtained 
    by the brute force algorithm; and the third must be a list of Collocation 
    instances obtained by the rotation-collocation algorithm. A dictionary of 
    "true_positive", "true_negative", "false_positive", "false_negative" counts is 
    produced."""

    #  Check input. 

    if not isinstance( collocations_bruteforce, list ) or not isinstance( collocations_rotation, list ): 
        raise collocationError( "InvalidArgument", "Both arguments must be lists" )

    if not isinstance( collocations_bruteforce[0], Collocation ): 
        raise collocationError( "InvalidArgument", "First argument must be a list of instances of Collocation" )

    if not isinstance( collocations_rotation[0], Collocation ): 
        raise collocationError( "InvalidArgument", "First argument must be a list of instances of Collocation" )

    n_total = occs.size
    n_in_bruteforce = len( collocations_bruteforce )
    n_in_rotation = len( collocations_rotation )
    n_in_both = len( collocations_bruteforce.intersection( collocations_rotation ) ) 

    ret = { 'true_positive': n_in_both, 
           'false_negative': n_in_bruteforce - n_in_both, 
           'false_positive': n_in_rotation - n_in_both, 
           'true_negative': n_total - ( n_in_bruteforce + n_in_rotation - n_in_both ) }

    return ret


class CollocationList( list ): 
    """A list of collocations. While inheriting the list class, it adds methods to 
    perform unions, intersections, and writing to output."""

    def __init__( self, input_list:list ): 
        """The input should be a list of instances of class Collocation."""

        if not isinstance( input_list, list ): 
                raise collocationError( "InvalidArgument", "Argument of CollocationList must be a list" )

        for element in input_list: 
            if not isinstance( element, Collocation ): 
                raise collocationError( "InvalidArgument", 
                        "All elements of the input list to CollocationList must be instances of Collocation" )

        super().__init__( input_list )

        return


    def union( self, union_list ): 
        """Return the union with the argument (instance of CollocationList)."""

        #  Check input. 

        if not isinstance( union_list, CollocationList ): 
            raise collocationError( "InvalidArgument", "Argument to CollocationList.union must be a CollocationList" )

        #  Create a dictionary corresponding to both lists. 

        dict1 = { c.occultation._data[0]['occid']: c for c in self }
        dict2 = { c.occultation._data[0]['occid']: c for c in union_list }

        #  Get union of "occid" keys. 

        keys = sorted( list( set( dict1.keys() ).union( set( dict2.keys() ) ) ) )

        #  Generate the result. 

        ret = CollocationList( [ dict1[k] for k in keys ] )

        #  Done. 

        return ret

    def intersection( self, union_list ): 
        """Return the intersection with the argument (instance of CollocationList)."""

        #  Check input. 

        if not isinstance( union_list, CollocationList ): 
            raise collocationError( "InvalidArgument", "Argument to CollocationList.intersection must be a CollocationList" )

        #  Create a dictionary corresponding to both lists. 

        dict1 = { c.occultation._data[0]['occid']: c for c in self }
        dict2 = { c.occultation._data[0]['occid']: c for c in union_list }

        #  Get intersection of "occid" keys. 

        keys = sorted( list( set( dict1.keys() ).intersection( set( dict2.keys() ) ) ) )

        #  Generate the result. 

        ret = CollocationList( [ dict1[k] for k in keys ] )

        #  Done. 

        return ret

    def write_to_netcdf( self, outputfile, author=None ): 
        """Write a list of collocations to an output NetCDF-4 file. The collocations 
        must be a list of instances of Collocation, and the file is the path of the 
        output file. Giving the author's name is optional."""

        #  Check input. 

        if not isinstance( outputfile, str ): 
            raise collocationError( "InvalidArgument", "The second argument must be a str, designating the path of the output file" )

        #  Create output file. 

        d = netCDF4.Dataset( outputfile, 'w' )

        #  Global attributes. 

        d.setncatts( { 
            'creation_time': Time().calendar("utc").isoformat(timespec="seconds")+"Z", 
            'file_type': "gnssro-nadirsounder-collocations" 
            } )

        if author is not None: 
            d.setncattr( "author", author )

        d.createDimension( "timestr", 19 )

        #  Loop over collocations. 

        for collocation in self: 

            #  Write occultation and sounder data. 

            occid = collocation.occultation._data[0]['occid']
            collocation_name = occid + "+" + \
                    collocation.nadir_satellite.satellite_name + "-" + \
                    collocation.nadir_satellite.instrument_name

            collocation_group = d.createGroup( collocation_name )
            collocation_group.setncattr( "status", collocation.status )

            longitude = collocation_group.createVariable( "longitude", "f4" )
            longitude.setncatts( { 'description': "Reference longitude of collocation, positive eastward", 'units': "degrees" } )

            latitude = collocation_group.createVariable( "latitude", "f4" )
            latitude.setncatts( { 'description': "Reference latitude of collocation, positive northward", 'units': "degrees" } )

            ctime = collocation_group.createVariable( "time", "c", dimensions=("timestr",) )
            ctime.setncatts( { 'description': "Reference time of collocation, UTC, ISO format string" } )

            if collocation.data is not None and collocation.status == "nominal": 

                occultation_group = collocation_group.createGroup( "occultation" )
                write_dataset_to_netcdf( collocation.data['occultation'], occultation_group )

                sounder_group = collocation_group.createGroup( "sounder" )
                write_dataset_to_netcdf( collocation.data['sounder'], sounder_group )

            longitude.assignValue( collocation.longitude )
            latitude.assignValue( collocation.latitude )
            ctime[:] = collocation.time.calendar("utc").isoformat(timespec="seconds")

        #  Done. 

        d.close()

        return

    ############################################################
    #  Magic methods. 
    ############################################################

    def __getitem__( self, ss ): 
        items = list( self ).__getitem__(ss)
        if isinstance( items, Collocation ): 
            ret = items
        else: 
            ret = CollocationList( items )

        return ret

    def __sort__( self, method="occtime" ): 

        if method == "occid": 
            sdict = { c.data['occid']: c for c in self } 

        elif method == "occtime": 
            sdict = {}
            for element in self: 
                time = c.data['occultation'].attrs['time']
                for ikeychar in range(26): 
                    key = time + chr( ord("a") + ikeychar )
                    if key not in sdict.keys(): 
                        break
                sdict.update( { key: element } )

        else: 
            raise collocationError( "InvalidArgument", f'Unrecognized method {method} to sort a CollocationList' )

        keys = sorted( list( sdict.keys() ) )
        ret = CollocationList( [ sdict[key] for key in keys ] )

        return ret


def write_dataset_to_netcdf( dataset, nc ): 
    """Write an xarray Dataset (dataset) to an open NetCDF file or group (nc)."""

    #  Check input. 

    if not isinstance( dataset, xarray.Dataset ): 
        raise collocationError( "InvalidArgument", "First argument must be an instance of xarray.Dataset" )

    if not isinstance( nc, netCDF4.Dataset ): 
        raise collocationError( "InvalidArgument", "Second argument must be an instance or child of netCDF4.Dataset" )

    #  Create dimensions. 

    for name, size in dataset.sizes.items(): 
        nc.createDimension( name, size )

    #  Create variables and their attributes. 

    variables = {}
    for vname, vobj in dataset.variables.items(): 
        v = nc.createVariable( vname, vobj.dtype, vobj.dims )
        v.setncatts( vobj.attrs )
        variables.update( { vname: v } )

    #  Create global attributes. 

    nc.setncatts( dataset.attrs )

    #  Write data values. 

    for vname, vobj in variables.items(): 
        vobj[:] = dataset.variables[vname].values

    #  Done. 

    return




"""
The `brute_force` module implements two different brute-force 
collocation-finding methods, which are used as base methods 
to generate truth data against which the results of the 
rotation-collocation methods can be compared.

Last updated: 9/8/2023 by Alex Meredith
"""

from .collocation import Collocation, CollocationList
from .constants_and_utils import calculate_radius_of_earth
from .awsro import get_occ_times
from .timestandards import Time
from datetime import datetime
import numpy as np
from tqdm import tqdm 


def brute_force( nadir_satellite_instrument, occs, time_tolerance, spatial_tolerance,
                use_sorting=True, progressbar=True ):
    """This function is a wrapper that performs brute-force colocation finding.
    It loops over all occultation/nadir-scanner sounding pairs and checks each
    pair for cotemporality and spatial overlap, in order to find colocations
    that overlap in both time and space.

    Arguments
    ------------
        nadir_satellite_instrument: NadirSatelliteInstrument
            An instance of a child of the NadirSatelliteInstrument base class. 
        nadir_scanner_geolocations: ScanMetadata
            An object containing geolocation information on the nadir-scan 
            footprints. It also contains methods to retrieve observation data
            given scan and footprint indices. 
        occs: awsgnssroutils.database.OccList
            An object containing geolocation information on RO soundings and 
            pointers to relevant data files where RO data can be retrieved. 
        time_tolerance: float
            Time tolerance for colocations in seconds
        spatial_tolerance: float
            Spatial tolerance for colocations in m

    Returns
    -----------
        A dictionary with the following keyword, value information: 

        status -> "success" or "fail"
        messages -> A list of mnemonic messages regarding performance
        comments -> A list of comments regarding performance
        data -> An instance of CollocationList containing collocations found
    """

    #  Initialize return dictionary. 

    ret = { 'status': None, 'messages': [], 'comments': [], 'data': None }

    #  Set parameters. 

    time_convention = "gps"

    #  Load nadir-scanner geolocation data. Define time range for nadir-scanner 
    #  sounding geolocation data. 

    occs_datetime64 = np.array( [ np.datetime64( t.calendar(time_convention).isoformat() ) for t in get_occ_times( occs ) ] )
    tmin = occs_datetime64.min() - np.timedelta64( time_tolerance, "s" )
    tmax = occs_datetime64.max() + np.timedelta64( time_tolerance, "s" )
    timerange = ( Time( **{ time_convention: tmin } ), Time( **{ time_convention: tmax } ) )

    nadir_scanner_geolocations = nadir_satellite_instrument.get_geolocations( timerange )

    #  Initialize and loop over occultation. 

    colocs = []

    if progressbar: 
        iterator = tqdm( range(occs.size), desc="Brute force" )
    else: 
        iterator = range(occs.size)

    for iocc in iterator: 

        occ = occs[iocc]
        resp = brute_force_sorted( nadir_scanner_geolocations, occ, 
                    time_tolerance, spatial_tolerance )

        if resp['status'] == "fail": 
            ret['status'] == "fail"
            ret['messages'] += resp['messages']
            ret['comments'] += resp['comments']
            return ret

        if resp['data'] is not None:

            iscan, ifootprint = resp['data']['iscan'], resp['data']['ifootprint']

            coloc = Collocation( occ, nadir_satellite_instrument, 
                          longitude = np.rad2deg( nadir_scanner_geolocations.longitudes[iscan,ifootprint] ), 
                          latitude = np.rad2deg( nadir_scanner_geolocations.latitudes[iscan,ifootprint] ), 
                          time = nadir_scanner_geolocations.mid_times[iscan], 
                          scan_metadata = nadir_scanner_geolocations, 
                          iscan = iscan, 
                          ifootprint = ifootprint )

            colocs.append( coloc )

    #  Finalize. 

    ret['status'] = "success"
    ret['data'] = CollocationList( colocs )

    return ret


def brute_force_sorted( nadir_scanner_geolocations, occ, 
        time_tolerance, spatial_tolerance ): 
    """This function uses a brute-force method to find nadir-scanner soundings
    colocated with a single radio occultation. This is the 'efficient' version
    of the brute-force method, which relies on sorting nadir-scanner data
    to shrink the search window.

    Arguments
    ------------
        nadir_scanner_geolocations: ScanMetadata
            An object containing geolocation information on the nadir-scan 
            footprints. It also contains methods to retrieve observation data
            given scan and footprint indices. 
        occ: awsgnssroutils.database.OccList
            An object containing geolocation information on just one RO 
            sounding with pointers to relevant data files where RO data can 
            be retrieved. 
        time_tolerance: float
            Time tolerance for colocations in seconds
        spatial_tolerance: float
            Spatial tolerance for colocations in m

    Returns
    -----------
        A dictionary with the following keyword, value information: 

        status -> "success" or "fail"
        messages -> A list of mnemonic messages regarding performance
        comments -> A list of comments regarding performance
        data -> None if no collocation found; a dictionary containing the 
                'iscan' and 'ifootprint' of the collocated sounding
    """

    #  Initialize return dictionary. 

    ret = { 'status': None, 'messages': [], 'comments': [], 'data': None }

    #  Get nadir-scanner scans that fall within time tolerance. 

    time_convention = "gps"

    #  Occultation geolocation processing. 

    occ_datetime64 = np.datetime64( get_occ_times(occ)[0].calendar(time_convention).isoformat() )
    occ_longitude = np.deg2rad( occ.values("longitude")[0] )
    occ_latitude = np.deg2rad( occ.values("latitude")[0] )

    #  Nadir scanner geolocation processing. 

    mid_times_datetime64 = np.array( [ np.datetime64( t.calendar(time_convention).isoformat() ) \
            for t in nadir_scanner_geolocations.mid_times ] )

    #  Time tolerance. 

    time_tolerance_timedelta64 = np.timedelta64( time_tolerance, "s" )

    indices = np.argwhere( np.logical_and( 
            mid_times_datetime64 >= occ_datetime64-time_tolerance_timedelta64, 
            mid_times_datetime64 <= occ_datetime64+time_tolerance_timedelta64 ) ).squeeze()

    nadir_lons = nadir_scanner_geolocations.longitudes[indices,:]
    nadir_lats = nadir_scanner_geolocations.latitudes[indices,:]

    #  Establish x,y,z coordinates of the nadir soundings. 

    p_nadir = np.array( [ np.cos(nadir_lons)*np.cos(nadir_lats), 
                             np.sin(nadir_lons)*np.cos(nadir_lats), 
                             np.sin(nadir_lats) ] ).transpose( (1,2,0) )

    #  Establish x,y,z coordinates of the occultation sounding. 

    p_occ = np.array( [ np.cos(occ_longitude)*np.cos(occ_latitude), 
                           np.sin(occ_longitude)*np.cos(occ_latitude), 
                           np.sin(occ_latitude) ] )

    #  Find angular distances to all nadir soundings (sorted in time). Find the 
    #  index of the minimum distance. 

    dist_ang = np.arccos( np.inner( p_nadir, p_occ ) )
    arg = dist_ang.argmin()
    iiscan = int( arg / dist_ang.shape[1] ) 
    iscan, ifootprint = indices[iiscan], arg % dist_ang.shape[1]

    #  Find distance of closest nadir sounding. Test against spatial tolerance. 

    r_e = calculate_radius_of_earth( 0.5 * ( occ_latitude + nadir_scanner_geolocations.latitudes[iscan,ifootprint] ) )

    #  Finalize. 

    ret['status'] = "success"

    if dist_ang[iiscan,ifootprint] <= spatial_tolerance/r_e * 1.0e-3: 
        ret['data'] = { 'iscan': iscan, 'ifootprint': ifootprint }
    else: 
        ret['data'] = None

    return ret




"""constants_and_utils.py

Author: Alex Meredith (ameredit@mit.edu), Stephen Leroy (sleroy@aer.com)
Version: 
Date: March 13, 2024

This module contains utilities useful for analyzing the geometry of 
nadir scanner soundings and radiation. It also contains attributes 
regarding the Earth's mass, shape, and spin."""


###  Imports #############################

import os
import numpy as np
import xarray as xr
import time
import erfa
import bisect
import netCDF4

### Earth-related constants ##############

Earth_equatorial_radius = 6378137.0/1000 #km
Earth_polar_radius = 6356752.3142/1000 #km
km_to_degree = 360/(2*np.pi*Earth_polar_radius)#degree/km
mu = 3.986004418e5 #standard gravitational parameter of Earth [km^3/s^2]
sec_to_sidereal_day = 1/86164.0905 #sidereal day/sec

### Physical constants ###################

planck_constant = 6.62607015e-34   # J/Hz
boltzmann_constant = 1.380649e-23  # J/K
speed_of_light = 2.99792458e8      # m/s

###  Parameters ##########################

HOME = os.path.expanduser( "~" )
defaults_file = os.path.join( HOME, ".collocation" )


### Utility functions ####################

def calculate_km_to_degree(lat_geodetic):
    """
    This function calculates the latitude-dependent conversion factor
    between km on the Earth's surface and degrees swept out from Earth's
    center.

    Parameters
    ------------
        lat_geodetic: float
            Geodetic latitude[radians]

    Returns
    --------
        km_to_degree: float
            Latitude-adjusted conversion factor between km and degrees
    """
    r_E = calculate_radius_of_earth(lat_geodetic)
    return np.rad2deg(1/r_E)


def calculate_radius_of_earth(lat_geodetic):
    """
    This function calculates the latitude-dependent radius of the Earth.

    Parameters
    ------------
        lat_geodetic: float
            Geodetic latitude[radians]

    Returns
    --------
        r_e: float
            Latitude-adjusted radius of Earth, in km
    """
    a = Earth_equatorial_radius
    b = Earth_polar_radius
    r_e = np.sqrt(((a**2*np.cos(lat_geodetic))**2 +
                   (b**2*np.sin(lat_geodetic))**2) /
                  ((a*np.cos(lat_geodetic))**2 +
                   (b*np.sin(lat_geodetic))**2))
    return r_e


def get_distance(lon_1, lat_1, lon_2, lat_2):
    """
    This function gets the distance between two lat/lon points assumed
    to be on the Earth's surface.

    Arguments
    -----------
        lon_1: float
            Longitude of first point, in radians
        lat_1: float
            Latitude of first point, in radians
        lon_2: float
            Longitude of second point, in radians
        lat_2: float
            Latitude of second point, in radians

    Returns
    ---------
        ang: float
            Distance in terms of angle between vectors from Earth's
            surface to each point, in radians
        dist: float
            Distance, in km
    """
    # Convert lat/lon to 3D vector
    vec_1 = np.array([np.cos(lon_1)*np.cos(lat_1),
                      np.sin(lon_1)*np.cos(lat_1),
                      np.sin(lat_1)])
    vec_2 = np.array([np.cos(lon_2)*np.cos(lat_2),
                      np.sin(lon_2)*np.cos(lat_2),
                      np.sin(lat_2)])

    # Get angle between vectors and convert to distance
    ang = np.arccos(np.dot(vec_1, vec_2))

    # Radius of Earth is latitude-dependent, approximate as avg between lats
    r_e = calculate_radius_of_earth((lat_1+lat_2)/2.0)

    return ang, r_e*ang


def constrain_angle_pair(ang1, ang2, num_full_revs=0):
    """
    Constrains angles such that -2pi < ang1 < 0 and ang1 <
    ang2 and such that |ang2 - ang1| < 2pi*(num_full_revs+1).

    Arguments
    -----------
        ang1: float
            angle in radians
        ang2: float
            angle in radians
        num_full_revs: int (optional)
            number of full revolutions between ang1 and ang2
            (e.g. if 0.75 revolutions between angles, num_full_revs=0)
    Returns
    --------
        ang1: float
            angle in radians constrained to be between -2pi and 0.
        ang2: float
            angle in radians constrained to be between ang1 and
            ang1 + 2pi * (num_full_revs + 1).
    """
    while ang1 < -2*np.pi:
        ang1 = ang1+2*np.pi
    while ang1 > 0:
        ang1 = ang1-2*np.pi
    while ang2 < ang1+2*np.pi*num_full_revs-np.pi:
        ang2 = ang2+2*np.pi
    while ang2 > ang1+2*np.pi*(num_full_revs+1)-np.pi:
        ang2 = ang2-2*np.pi
    return ang1, ang2


def constrain_to_pi_range(ang):
    """
    Constrains an angle (in radians) to between -pi and +pi.

    Parameters
    -------------
        ang: float
            angle in radians
    Returns
    ---------
        ang: float
            angle in radians constrained to be between -pi and pi.
    """
    while ang < -np.pi:
        ang = ang+2*np.pi
    while ang > np.pi:
        ang = ang-2*np.pi
    return ang


def masked_dataarray( data, fill_value, attrs={}, **kwargs ): 
    """Create a masked version of an xarray.DataArray. If the input data is a numpy masked array, 
    then the masked slots are filled with fill_value and the fill value is written into the 
    DataArray attributes as _FillValue. An xarray.DataArray is returned."""

    #  Check input. 

    if isinstance( data, np.ma.core.MaskedArray ): 

        xdata = data.data
        xfill = data.dtype.type( fill_value )

        if len( xdata.shape ) == 0: 
            if data.mask: 
                xdata = xfill
        else: 
            i = np.argwhere( data.mask ).flatten()
            xdata[i] = xfill

        attrs.update( { '_FillValue': xfill } )
        ret = xr.DataArray( xdata, attrs=attrs, **kwargs )

    elif isinstance( data, np.ndarray ): 

        ret = xr.DataArray( data, attrs=attrs, **kwargs )

    return ret


def planck_blackbody( frequency, temperature ): 
    """Planck blackbody radiation. The arguments can be scalars or 
    numpy.ndarray's. Frequency is Hz, temperature in K. Output 
    units are W m**-2 Hz**-1 steradian**-1."""

    ret = 2 * planck_constant * frequency**3 / speed_of_light**2 / \
            ( np.exp( ( planck_constant * frequency ) / ( boltzmann_constant * temperature ) ) - 1 )

    return ret





"""Utilities for reading Celestrak TLE data. All TLE data must be 
obtained from celestrak.org as special requests for GP elements: 

    https://celestrak.org/NORAD/archives/request.php

Once the data are received by email, the files must be written into 
a path with data_root as the root. Data files can be arranged in 
subdirectories at the user's discretion, but the file names must 
satisfy the regular expression prescribed by 
"^sat{norad_number:09d}.*\.txt$" where norad_number is an integer 
that defines the NORAD catalogue number of the satellite whose TLE 
data are in the file. THE DATA FOR ONLY ONE SATELLITE IN EACH FILE, 
and these are definitely 2-line TLE files.

Example usage: 

    > from datetime import datetime
    > cs = Celestrak( "/home/user/Data/celestrak_tles" )
    > metopa = cs.satellite( 29499 )
    > tles = metopa.select( ( datetime(2009,8,1), datetime(2009,9,1) ), extend=(1,1) )

which retries all TLE for Metop-A for the month of August, 2009, including 
the last record for July, 2009, and the first for September, 2009. 
    """

#  Imports. 

import os, re, json, stat
import numpy as np
from .timestandards import Time, Calendar
from .constants_and_utils import defaults_file

#  Definitions. 

root_path_variable = "celestrak_data_root"
celestrak_time_convention = "utc"
gps0 = Time( gps=0 )

#  Table of satellites. 

Satellites = [ { 'name': "Metop-A", 'norad_number': 29499 }, 
              { 'name': "Metop-B", 'norad_number': 38771 }, 
              { 'name': "Metop-C", 'norad_number': 43689 }, 
              { 'name': "Suomi-NPP", 'norad_number': 37849 }, 
              { 'name': "JPSS-1", 'norad_number': 43013 }, 
              { 'name': "JPSS-2", 'norad_number': 54234 } ]


#  Exception handling. 

class Error( Exception ): 
    pass

class celestrakError( Error ): 
    def __init__( self, message, comment ): 
        self.message = message
        self.comment = comment


def setdefaults( root_path ):
    """This method sets the default root path for Celestrak TLE data."""

    if os.path.exists( defaults_file ):
        with open( defaults_file, 'r' ) as f:
            defaults = json.load( f )
        defaults.update( { root_path_variable: root_path } ) 
    else: 
        defaults = {}

    with open( defaults_file, 'w' ) as f:  
        json.dump( defaults, f, indent="  " ) 
    os.chmod( defaults_file, stat.S_IRUSR | stat.S_IWUSR )

    #  Done. 

    return





class CelestrakSatellite(): 
    """This class contains the TLE data, as obtained from Celestrak, for one 
    satellite."""

    def __init__( self, Celestrak_instance, norad_number ): 
        """Create an instance of class CelestrakSatellite. It will contain the 
        TLE data for a satellite identified by its NORAD catalogue number 
        (norad_number). The first argument is the instance of class Celestrak 
        from which this class is drawn."""

        if not isinstance( Celestrak_instance, Celestrak ): 
            raise celestrakError( "InvalidArgument", "First argument to " + \
                    "CelestrakSatellite not an instance of class Celestrak" )

        self.celestrak = Celestrak_instance
        norad_number_string = f'{norad_number:05d}'
        data = {}

        if norad_number_string not in self.celestrak.catalogue.keys(): 
            print( f'NORAD number {norad_number_string} not in Celestrak local database' )
            self.status = "fail"
            return

        for file in self.celestrak.catalogue[norad_number_string]: 

            f = open( file, 'r' )
            lines = f.readlines()
            f.close()

            if len(lines) % 2 == 1: 
                raise celestrakError( "InvalidData", f'file {file} has an odd number of lines' )

            for iline in range(0,len(lines),2): 

                tle_line1 = lines[iline]
                tle_line2 = lines[iline+1]

                if tle_line1[0] != "1": 
                    raise celestrakError( "InvalidData", 
                            f'Line {iline+1} in file {file} should be the first line of a TLE but is not' )

                if tle_line2[0] != "2": 
                    raise celestrakError( "InvalidData", 
                            f'Line {iline+2} in file {file} should be the second line of a TLE but is not' )

                #  Get the time of the record. 

                year = 2000 + int( tle_line1[18:20] )
                doy = int( tle_line1[20:23] )
                fday = float( tle_line1[23:32] )

                time = Time( **{ celestrak_time_convention: Calendar(year,1,1) } ) + 86400 * ( doy + fday - 1 )
                rec = { 'time': time, 'tle': ( tle_line1, tle_line2 ) }
                data.update( { time.calendar(celestrak_time_convention).isoformat(timespec="seconds"): rec } )

        #  Sort the records. 

        sorted_key_list = sorted( list( data.keys() ) )

        #  Create data contents. 

        self.nrecs = len( sorted_key_list )
        self.datetimes = [ data[key]['time'] for key in sorted_key_list ]
        self.tles = [ data[key]['tle'] for key in sorted_key_list ]

        return

    def select( self, timerange, extend=None ): 
        """Generate a list of TLEs. Each element of output will be a 2-tuple containing 
        the two lines of a TLE. All TLEs in the list must fall within a range of 
        datetimes defined by the two-element tuple/list datetimerange. 

        Arguments: 

        timerange       A two-element tuple/list containing two instances of timestandards.Time. 
                        The first element defines the start time of the TLE list to be 
                        generated; the second defines the end time of the TLE list. 

        extend          A two-element tuple/list of integers defining the how far to extend 
                        the list of TLEs before the datetimerange and after the datetimerange. 

        Example: 

        data = CelestrakSatellite.select( 
                timerange = [ timestandards.Time( utc=timestandards.Calendar(2023,6,7) ), 
                                timestandards.Time( utc=timestandards.Calendar(2023,6,8) ) ], 
                extend = [0,1] )

        The statement generates a list of TLEs for the satellite defined by CelestrakSatllite
        for the day of June 7, 2023, and appends to the list the first TLE for the succeeding 
        day."""

        if not isinstance(timerange,list) and not isinstance(timerange,tuple): 
            raise celestrakError( "InvalidArgument", 'timerange must be a tuple or list' )

        if len( timerange ) != 2: 
            raise celestrakError( "InvalidArgument", 'timerange must containing two elements' )

        if not isinstance(timerange[0],Time) or not isinstance(timerange[1],Time): 
            raise celestrakError( "InvalidArgument", 'timerange must consist of timestandards.Time instances' )

        if extend is not None: 

            if not isinstance(extend,list) and not isinstance(extend,tuple): 
                raise celestrakError( "InvalidArgument", 'extend must be a tuple or list' )

            if len( extend ) != 2: 
                raise celestrakError( "InvalidArgument", 'extend must containing two elements' )

            if not isinstance(extend[0],int) or not isinstance(extend[1],int): 
                raise celestrakError( "InvalidArgument", 'extend must consist of two int' )

        #  Extract elements. 

        dtr = np.array( [ dt - gps0 for dt in timerange ] )
        datetimes_gps = np.array( [ dt - gps0 for dt in self.datetimes ] )
        ii = np.argwhere( np.logical_and( dtr[0] <= datetimes_gps, datetimes_gps < dtr[1] ) ).squeeze()

        #  Extend? 

        if extend is None: 
            imin = ii[0]
            imax = ii[-1]
        else: 
            imin = max( [ 0, ii[0]-extend[0] ] )
            imax = min( [ self.nrecs-1, ii[-1]+extend[1] ] )

        #  Extract TLEs. 

        ret = [ self.tles[i] for i in range(imin,imax+1) ]

        #  Done. 

        return ret


    def nearest( self, time ): 
        """Get the TLE that is nearest in time (instance of timestandards.Time). The two lines of the 
        TLE are returned as a 2-tuple if successful. If unsuccessful, None is returned."""

        ret = { 'status': None, 'messages': [], 'comments': [] }

        if not isinstance( time, Time ): 
            ret['messages'].append( "InvalidArgument" )
            ret['comments'].append( f'The argument to Celestrak.nearest must be an instance of timestandards.Time' )
            ret['status'] = "fail"
            return ret

        datetimes_gps = np.array( [ dt-gps0 for dt in self.datetimes ] )
        time_gps = time - gps0

        i = np.argmin( np.abs( datetimes_gps - time_gps ) ) 

        #  Check for valid return value. The nearest time must be better than 12 hours. 

        if np.abs( datetimes_gps[i] - time_gps ) < 12 * 3600: 
            ret.update( { 'data': self.tles[i] } )
            ret['status'] = "success"
        else: 
            ret['status'] = "fail"

        return ret



class Celestrak(): 
    """This is a portal to handling TLE data obtained from Celestrak."""

    def __init__( self ): 
        """Create a portal to access TLE downloaded from celestrak.org. Data must be 
        strict two-line element files, with each file containing data for only one 
        satellite."""

        with open( defaults_file, 'r' ) as f: 
            defaults = json.load( f )
            self.data_root = defaults[root_path_variable]

        if not os.path.isdir( self.data_root ): 
            raise celestrakError( "PathNotFound", f'"{root_path_variable} does not exist or is not a directory' )

        #  Scan directory hierarchy looking for data files. 

        self.catalogue = {}

        for root, dirs, files in os.walk( self.data_root ): 

            files.sort()
            dirs.sort()

            for file in files: 
                m = re.search( "^sat(\d{9}).*\.txt$", file )
                if not m: continue

                #  Read first line of the file to get the satellite NORAD number. 

                file_path = os.path.join( root, file )
                f = open( file_path, 'r' )
                line = f.readline()
                norad_number_string = line[2:7]
                f.close()

                #  Catalog the file. 

                if norad_number_string not in self.catalogue.keys(): 
                    self.catalogue.update( { norad_number_string: [] } )
                self.catalogue[norad_number_string].append( file_path )

            pass 

        #  Done. 

        pass

    def satellite( self, generic ): 
        """Create an instance of class CelestrakSatellite given either a 
        NORAD catalogue number (int) or the name of the satellite (str) as 
        given in the table Satellites."""

        #  Define the NORAD catalogue number of the satellite. 
        
        norad_number = 0
        
        if isinstance( generic, int ): 
            norad_number = generic
        elif isinstance( generic, str ): 
            s = [ sat for sat in Satellites if sat['name'] == generic ]
            if len(s) == 1: 
                norad_number = s[0]['norad_number']
                
        #  Convert to string. 
        
        if norad_number != 0: 
            norad_number_string = f'{norad_number:05d}'
        else: 
            norad_number_string = ""
        
        #  Generate the Celestrak satellite, load TLE data. 
        
        if norad_number_string not in self.catalogue.keys(): 
            ret = None
        else: 
            ret = CelestrakSatellite( self, norad_number )

        return ret



if __name__ == "__main__": 

    data_dir = "../../Data/TLEs"
    cs = Celestrak( data_dir )

    #  Select a satellite. 

    metopb = cs.satellite(43689)
    tle_data = metopb.select( ( datetime(2022,8,1), datetime(2022,9,1) ), extend=(1,1) )

    pass

